#include "cpudetect.h"
#include "debug.h"

struct sCPUFeatures gCpuFeatures;

#ifdef ARCH_X86

#   include <stdio.h>
#   include <string.h>

#   ifdef __NetBSD__
#       include <sys/param.h>
#       include <sys/sysctl.h>
#       include <machine/cpu.h>
#   endif

#   ifdef __FreeBSD__
#       include <sys/types.h>
#       include <sys/sysctl.h>
#   endif

#   ifdef __linux__
#       include <signal.h>
#   endif

#endif // #ifdef ARCH_X86

/* Thanks to the FreeBSD project for some of this cpuid code, and 
 * help understanding how to use it.  Thanks to the Mesa 
 * team for SSE support detection and more cpu detect code.
 */
/* I believe this code works.  However, it has only been used on a PII and PIII */
#ifdef ARCH_X86
static void check_os_katmai_support( void );

/* Check whether the cpuid instructions is supported by the CPU */
static int has_cpuid()
{
    int a, c;

    /* code from libavcodec: */
    __asm__ __volatile__ (
        /* See if CPUID instruction is supported ... */
        /* ... Get copies of EFLAGS into eax and ecx */
        "pushf\n\t"
        "popl %0\n\t"
        "movl %0, %1\n\t"
                          
        /* ... Toggle the ID bit in one copy and store */
        /*     to the EFLAGS reg */
        "xorl $0x200000, %0\n\t"
        "push %0\n\t"
        "popf\n\t"
                         
        /* ... Get the (hopefully modified) EFLAGS */
        "pushf\n\t"
        "popl %0\n\t"
        : "=a" (a), "=c" (c)
        :
        : "cc" 
    );

    return (a!=c);
}

/* Perform the cpuid call and get the info */
static void
do_cpuid(unsigned int ax, unsigned int *p)
{
    /* code from libavcodec: */
    __asm __volatile
	("movl %%ebx, %%esi\n\t"
         "cpuid\n\t"
         "xchgl %%ebx, %%esi"
         : "=a" (p[0]), "=S" (p[1]), 
           "=c" (p[2]), "=d" (p[3])
         : "0" (ax));
}

int get_cpu_vendor(unsigned int regs[], unsigned int regs2[])
{
    #define CPUID_EXTFAMILY ((regs2[0] >> 20)&0xFF) /* 27..20 */
    #define CPUID_EXTMODEL  ((regs2[0] >> 16)&0x0F) /* 19..16 */
    #define CPUID_TYPE	    ((regs2[0] >> 12)&0x04) /* 13..12 */
    #define CPUID_FAMILY    ((regs2[0] >>  8)&0x0F) /* 11..08 */
    #define CPUID_MODEL	    ((regs2[0] >>  4)&0x0F) /* 07..04 */
    #define CPUID_STEPPING  ((regs2[0] >>  0)&0x0F) /* 03..00 */

    #include "x86cputable.h" /* get cpuname and cpuvendors */
    char vendor[32];
    int i, cpuVendor = eVendorIntel;

    snprintf(vendor,31,"%.4s%.4s%.4s",
            (char*)(regs+1),(char*)(regs+3),(char*)(regs+2));

    for(i=0; i<MAX_VENDORS; i++)
    {
	if(!strcmp(cpuvendors[i].string,vendor))
        {
            cpuVendor = cpuvendors[i].vendor;
            if(cpuname[i][CPUID_FAMILY][CPUID_MODEL])
            {
		EQ_DEBUG("%s %s",cpuvendors[i].name,cpuname[i][CPUID_FAMILY][CPUID_MODEL]);
            }
            break;
	}
        /*
        else 
        {
            snprintf(namebuf,255,"unknown %s %d. Generation CPU",cpuvendors[i].name,CPUID_FAMILY); 
            EQ_DEBUG("unknown %s CPU:\n",cpuvendors[i].name);
	    EQ_DEBUG("Vendor:   %s\n",cpuvendors[i].string);
            EQ_DEBUG("Type:     %d\n",CPUID_TYPE);
	    EQ_DEBUG("Family:   %d (ext: %d)\n",CPUID_FAMILY,CPUID_EXTFAMILY);
	    EQ_DEBUG("Model:    %d (ext: %d)\n",CPUID_MODEL,CPUID_EXTMODEL);
	    EQ_DEBUG("Stepping: %d\n",CPUID_STEPPING);
	    EQ_DEBUG("Please send the above info along with the exact CPU name"
			       "to the developers, so we can add it to the list!\n");
	} */
    }

#undef CPUID_EXTFAMILY
#undef CPUID_EXTMODEL
#undef CPUID_TYPE
#undef CPUID_FAMILY
#undef CPUID_MODEL
#undef CPUID_STEPPING
    return cpuVendor;
}


/* Interface to the world */
/* features must be provided by the client */
void get_cpu_features(struct sCPUFeatures *features)
{
    unsigned int regs[4];
    unsigned int regs2[4];

    memset(features, 0, sizeof(*features));
    features->isX86=1;

    if (!has_cpuid()) 
    {
        EQ_DEBUG("CPUID not supported!??? (maybe an old 486?)\n");
        return;
    }

    do_cpuid(0x00000000, regs); // get _max_ cpuid level and vendor name
    EQ_DEBUG("CPU vendor name: %.4s%.4s%.4s  max cpuid level: %d\n",
	    (char*) (regs+1),(char*) (regs+3),(char*) (regs+2), regs[0]);
    
    if (regs[0]>=0x00000001)
    {
        unsigned cl_size = 32; // Default

        do_cpuid(0x00000001, regs2);

        features->cpuVendor = get_cpu_vendor(regs, regs2);

	features->cpuType=(regs2[0] >> 8)&0xf;
	if(features->cpuType==0xf)
        {
	    // use extended family (P4, IA64)
	    features->cpuType=8+((regs2[0]>>20)&255);
	}
	features->cpuStepping=regs2[0] & 0xf;
	EQ_DEBUG("(Family: %d, Stepping: %d)\n",
		    features->cpuType, features->cpuStepping);

	// general feature flags:
	features->hasMMX  = (regs2[3] & (1 << 23 )) >> 23; // 0x0800000
	features->hasSSE  = (regs2[3] & (1 << 25 )) >> 25; // 0x2000000
	features->hasSSE2 = (regs2[3] & (1 << 26 )) >> 26; // 0x4000000
	features->hasMMX2 = features->hasSSE; // SSE cpus supports mmxext too
        
	cl_size = ((regs2[1] >> 8) & 0xFF)*8;
	if(cl_size) 
            features->clSize = cl_size;
    }

    do_cpuid(0x80000000, regs);
    if (regs[0]>=0x80000001) 
    {
	EQ_DEBUG("extended cpuid-level: %d\n",regs[0]&0x7FFFFFFF);
        do_cpuid(0x80000001, regs2);
	features->hasMMX  |= (regs2[3] & (1 << 23 )) >> 23; // 0x0800000
	features->hasMMX2 |= (regs2[3] & (1 << 22 )) >> 22; // 0x400000
	features->has3DNow    = (regs2[3] & (1 << 31 )) >> 31; //0x80000000
	features->has3DNowExt = (regs2[3] & (1 << 30 )) >> 30;
    }

    if(regs[0]>=0x80000006)
    {
	do_cpuid(0x80000006, regs2);
	EQ_DEBUG("extended cache-info: %d\n",regs2[2]&0x7FFFFFFF);
	features->clSize  = regs2[2] & 0xFF;
    }

    EQ_DEBUG("Detected cache-line size is %u bytes\n",features->clSize);

    /* FIXME: Does SSE2 need more OS support, too? */
#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__)
    if (features->hasSSE)
	check_os_katmai_support();
    if (!features->hasSSE)
	features->hasSSE2 = 0;
#else
    features->hasSSE=0;
    features->hasSSE2 = 0;
#endif

#ifndef USE_MMX
    if(features->hasMMX)
        EQ_DEBUG("MMX supported but disabled\n");
    features->hasMMX=0;
#endif
#ifndef USE_MMX2
    if(features->hasMMX2) 
        EQ_DEBUG("MMX2 supported but disabled\n");
    features->hasMMX2=0;
#endif
#ifndef USE_SSE
    if(features->hasSSE) 
        EQ_DEBUG("SSE supported but disabled\n");
    features->hasSSE=0;
#endif
#ifndef USE_SSE2
    if(features->hasSSE2) 
        EQ_DEBUG("SSE2 supported but disabled\n");
    features->hasSSE2=0;
#endif
#ifndef USE_3DNOW
    if(features->has3DNow) 
        EQ_DEBUG("3DNow supported but disabled\n");
    features->has3DNow=0;
#endif
#ifndef USEE_3DNOWEX
    if(features->has3DNowExt) 
        EQ_DEBUG("3DNowExt supported but disabled\n");
    features->has3DNowExt=0;
#endif
}

#if defined(__linux__) && defined(_POSIX_SOURCE) && defined(X86_FXSR_MAGIC)
static void sigill_handler_sse( int signal, struct sigcontext sc )
{
   EQ_DEBUG("SIGILL, " );

   /* Both the "xorps %%xmm0,%%xmm0" and "divps %xmm0,%%xmm1"
    * instructions are 3 bytes long.  We must increment the instruction
    * pointer manually to avoid repeated execution of the offending
    * instruction.
    *
    * If the SIGILL is caused by a divide-by-zero when unmasked
    * exceptions aren't supported, the SIMD FPU status and control
    * word will be restored at the end of the test, so we don't need
    * to worry about doing it here.  Besides, we may not be able to...
    */
   sc.eip += 3;

   gCpuFeatures.hasSSE=0;
}

static void sigfpe_handler_sse( int signal, struct sigcontext sc )
{
   EQ_DEBUG("SIGFPE, " );

   if ( sc.fpstate->magic != 0xffff ) {
      /* Our signal context has the extended FPU state, so reset the
       * divide-by-zero exception mask and clear the divide-by-zero
       * exception bit.
       */
      sc.fpstate->mxcsr |= 0x00000200;
      sc.fpstate->mxcsr &= 0xfffffffb;
   } else {
      /* If we ever get here, we're completely hosed.
       */
      EQ_DEBUG("\n\nSSE enabling test failed badly!" );
   }
}
#endif /* __linux__ && _POSIX_SOURCE && X86_FXSR_MAGIC */

/* If we're running on a processor that can do SSE, let's see if we
 * are allowed to or not.  This will catch 2.4.0 or later kernels that
 * haven't been configured for a Pentium III but are running on one,
 * and RedHat patched 2.2 kernels that have broken exception handling
 * support for user space apps that do SSE.
 */
static void check_os_katmai_support( void )
{
#if defined(__FreeBSD__)
   int has_sse=0, ret;
   size_t len=sizeof(has_sse);

   ret = sysctlbyname("hw.instruction_sse", &has_sse, &len, NULL, 0);
   if (ret || !has_sse)
      gCpuFeatures.hasSSE=0;

#elif defined(__NetBSD__)
#   if __NetBSD_Version__ >= 105250000
   int has_sse, has_sse2, ret, mib[2];
   size_t varlen;

   mib[0] = CTL_MACHDEP;
   mib[1] = CPU_SSE;
   varlen = sizeof(has_sse);

   EQ_DEBUG("Testing OS support for SSE... " );
   ret = sysctl(mib, 2, &has_sse, &varlen, NULL, 0);
   if (ret < 0 || !has_sse) 
   {
      gCpuFeatures.hasSSE=0;
      EQ_DEBUG("no!\n" );
   } 
   else 
   {
      gCpuFeatures.hasSSE=1;
      EQ_DEBUG("yes!\n" );
   }

   mib[1] = CPU_SSE2;
   varlen = sizeof(has_sse2);
   EQ_DEBUG("Testing OS support for SSE2... " );
   ret = sysctl(mib, 2, &has_sse2, &varlen, NULL, 0);
   if (ret < 0 || !has_sse2) 
   {
      gCpuFeatures.hasSSE2=0;
      EQ_DEBUG("no!\n" );
   } 
   else 
   {
      gCpuFeatures.hasSSE2=1;
      EQ_DEBUG("yes!\n" );
   }
#   else
   gCpuFeatures.hasSSE = 0;
   EQ_DEBUG("No OS support for SSE, disabling to be safe.\n" );
#   endif
#elif defined(__linux__)
#   if defined(_POSIX_SOURCE) && defined(X86_FXSR_MAGIC)
   struct sigaction saved_sigill;
   struct sigaction saved_sigfpe;

   /* Save the original signal handlers.
    */
   sigaction( SIGILL, NULL, &saved_sigill );
   sigaction( SIGFPE, NULL, &saved_sigfpe );

   signal( SIGILL, (void (*)(int))sigill_handler_sse );
   signal( SIGFPE, (void (*)(int))sigfpe_handler_sse );

   /* Emulate test for OSFXSR in CR4.  The OS will set this bit if it
    * supports the extended FPU save and restore required for SSE.  If
    * we execute an SSE instruction on a PIII and get a SIGILL, the OS
    * doesn't support Streaming SIMD Exceptions, even if the processor
    * does.
    */
   if ( gCpuFeatures.hasSSE ) 
   {
      EQ_DEBUG("Testing OS support for SSE... " );

//      __asm __volatile ("xorps %%xmm0, %%xmm0");
      __asm __volatile ("xorps %xmm0, %xmm0");

      if ( gCpuFeatures.hasSSE ) 
      {
	 EQ_DEBUG("yes.\n" );
      }
      else 
      {
	 EQ_DEBUG("no!\n" );
      }
   }

   /* Emulate test for OSXMMEXCPT in CR4.  The OS will set this bit if
    * it supports unmasked SIMD FPU exceptions.  If we unmask the
    * exceptions, do a SIMD divide-by-zero and get a SIGILL, the OS
    * doesn't support unmasked SIMD FPU exceptions.  If we get a SIGFPE
    * as expected, we're okay but we need to clean up after it.
    *
    * Are we being too stringent in our requirement that the OS support
    * unmasked exceptions?  Certain RedHat 2.2 kernels enable SSE by
    * setting CR4.OSFXSR but don't support unmasked exceptions.  Win98
    * doesn't even support them.  We at least know the user-space SSE
    * support is good in kernels that do support unmasked exceptions,
    * and therefore to be safe I'm going to leave this test in here.
    */
   if ( gCpuFeatures.hasSSE ) 
   {
      EQ_DEBUG("Testing OS support for SSE unmasked exceptions... " );

//      test_os_katmai_exception_support();

      if ( gCpuFeatures.hasSSE ) 
      {
	 EQ_DEBUG("yes.\n" );
      } 
      else 
      {
	 EQ_DEBUG("no!\n" );
      }
   }

   /* Restore the original signal handlers.
    */
   sigaction( SIGILL, &saved_sigill, NULL );
   sigaction( SIGFPE, &saved_sigfpe, NULL );

   /* If we've gotten to here and the XMM CPUID bit is still set, we're
    * safe to go ahead and hook out the SSE code throughout Mesa.
    */
   if ( gCpuFeatures.hasSSE ) 
   {
      EQ_DEBUG("Tests of OS support for SSE passed.\n" );
   }
   else 
   {
      EQ_DEBUG("Tests of OS support for SSE failed!\n" );
   }
#   else
   /* We can't use POSIX signal handling to test the availability of
    * SSE, so we disable it by default.
    */
   EQ_DEBUG("Cannot test OS support for SSE, disabling to be safe.\n" );
   gCpuFeatures.hasSSE=0;
#   endif /* _POSIX_SOURCE && X86_FXSR_MAGIC */
# else
   /* Do nothing on other platforms for now.
    */
   EQ_DEBUG("Cannot test OS support for SSE, leaving disabled.\n" );
   gCpuFeatures.hasSSE=0;
# endif /* __linux__ */
}
#else /* ARCH_X86 */

# ifdef SYS_DARWIN
#   include <sys/sysctl.h>
# else
#   include <signal.h>
#   include <setjmp.h>

static sigjmp_buf jmpbuf;
static volatile sig_atomic_t canjump = 0;

static void sigill_handler (int sig)
{
    if (!canjump) 
    {
        signal (sig, SIG_DFL);
        raise (sig);
    }
    
    canjump = 0;
    siglongjmp (jmpbuf, 1);
}
# endif

void get_cpu_features(struct sCPUFeatures *features)
{
	features->cpuType=0;
	features->cpuStepping=0;
	features->hasMMX=0;
	features->hasMMX2=0;
	features->has3DNow=0;
	features->has3DNowExt=0;
	features->hasSSE=0;
	features->hasSSE2=0;
	features->isX86=0;
	features->hasAltiVec = 0;
# ifdef HAVE_ALTIVEC   
#  ifdef SYS_DARWIN   
/*
  rip-off from ffmpeg altivec detection code.
  this code also appears on Apple's AltiVec pages.
 */
        {
                int sels[2] = {CTL_HW, HW_VECTORUNIT};
                int has_vu = 0;
                size_t len = sizeof(has_vu);
                int err;

                err = sysctl(sels, 2, &has_vu, &len, NULL, 0);   

                if (err == 0)
                        if (has_vu != 0)
                                features->hasAltiVec = 1;
        }
#  else /* SYS_DARWIN */
/* no Darwin, do it the brute-force way */
/* this is borrowed from the libmpeg2 library */
        {
          signal (SIGILL, sigill_handler);
          if (sigsetjmp (jmpbuf, 1)) {
            signal (SIGILL, SIG_DFL);
          } else {
            canjump = 1;
            
            asm volatile ("mtspr 256, %0\n\t"
                          "vand %%v0, %%v0, %%v0"
                          :
                          : "r" (-1));
            
            signal (SIGILL, SIG_DFL);
            features->hasAltiVec = 1;
          }
        }
#  endif /* SYS_DARWIN */
        EQ_DEBUG("AltiVec %sfound\n", (features->hasAltiVec ? "" : "not "));
# endif /* HAVE_ALTIVEC */
}
#endif /* !ARCH_X86 */
